# python3-opsi
# This code is part of the opsi.org project
# http://opsi.org
#
# Copyright (c) uib gmbh (www.uib.de)
# This sourcecode is owned by the uib gmbh, D55118 Mainz, Germany
#
# This source code is part of co-funding project, and so are
# ALL RIGHTS RESEVERD right now
# http://uib.de/en/opsi_cofunding
# Once the co-funding is finished it will be released as full open source


if not bootimageVersion or (int(bootimageVersion) < 20160517):
    raise Exception(u"This product requires opsi-linux-bootimage version >= 20160517")
try:
    if not backend.backend_info()['opsiVersion'].startswith('4.'):
        raise ValueError('No opsi 4.x backend present')
except Exception:
    raise Exception(u"This is a opsi 4.x product")

isOpsi43 = False
if backend.backend_info()['opsiVersion'].startswith('4.3'):
    isOpsi43 = True

if isOpsi43:
    show_message(u"Installing Product: %s" % (productId), 10)

try:
    import ConfigParser
except:
    import configparser
from OPSI.Util.File import IniFile
from OPSI.Backend.BackendManager import BackendManager
from OPSI.Backend.JSONRPC import JSONRPCBackend
from OPSI.Util import md5sum
from hashlib import md5
try:
    from twisted.conch.ssh import keys
except:
    pass

#include opsi setup library
depot.copy(SCRIPT_PATH+'/opsisetuplib.py','/tmp')
exec(open('/tmp/opsisetuplib.py').read())

checkModules('local_imaging')

logger.notice(u"##############################################################")
initOpsiSetupLib()
logger.notice(u"opsi-local-image-restore.py")
logger.notice(u"##############################################################")
productVersion =''
packageVersion =''
for i in range(len(productOnDepots)):
    if (productOnDepots[i]['productId'] == productId):
        productVersion = productOnDepots[i]['productVersion']
        packageVersion = productOnDepots[i]['packageVersion']
    break
if isOpsi43:
    show_message(u"Installing Product: %s_%s-%s on %s" % (productId,productVersion,packageVersion, clientId), 15)
logger.notice(u"Installing Product: %s_%s-%s on %s" % (productId,productVersion,packageVersion,clientId))
scriptMessageSubject.setMessage(u"Installing Product: %s_%s-%s on %s" % (productId,productVersion,packageVersion,clientId))
logger.notice(u"##############################################################")

logger.notice(u"Using built in partclone")


def restoreImage(diskDevice, partition, imageFile, progressSubject=None, imageType = None, filesystem = "ntfs", useGpt = False):
    if not useGpt:
        for hook in hooks:
            (partition, imageFile, progressSubject) = hook.pre_Harddisk_restoreImage(disk, partition, imageFile, progressSubject)

    try:
        partition = forceInt(partition)
        imageFile = forceUnicode(imageFile)
        logger.notice(u"restoreImage(%s, %s, %s, %s, %s, %s, %s)" % (diskDevice, partition, imageFile, progressSubject, imageType, filesystem, useGpt))
        image = None
        fs = None

        pipe = u''
        if imageFile.endswith(u'|'):
            pipe = imageFile
            imageFile = u'-'

        if not imageType:
            try:
                head = u''
                if pipe:
                    proc = subprocess.Popen(
                            pipe[:-1] + u" 2>/dev/null",
                            shell   = True,
                            stdin   = subprocess.PIPE,
                            stdout  = subprocess.PIPE,
                            stderr  = None,
                    )
                    pid = proc.pid

                    head = proc.stdout.read(128)
                    logger.debug(u"Read 128 Bytes from pipe '%s': %s" % (pipe, head.decode('ascii', 'replace')))

                    proc.stdout.close()
                    proc.stdin.close()

                    while( proc.poll() == None ):
                        pids = os.listdir("/proc")
                        for p in pids:
                            if not os.path.exists( os.path.join("/proc", p, "status") ):
                                continue
                            f = open( os.path.join("/proc", p, "status") )
                            for line in f.readlines():
                                if line.startswith("PPid:"):
                                    ppid = line.split()[1].strip()
                                    if (ppid == str(pid)):
                                        logger.info(u"Killing process %s" % p)
                                        os.kill(int(p), SIGKILL)

                        logger.info(u"Killing process %s" % pid)
                        os.kill(pid, SIGKILL)
                        time.sleep(1)
                else:
                    image = open(imageFile, 'r')
                    head = image.read(128)
                    logger.debug(u"Read 128 Bytes from file '%s': %s" % (imageFile, head.decode('ascii', 'replace')))
                    image.close()

                if   (head.find('ntfsclone-image') != -1):
                    logger.notice(u"Image type is ntfsclone")
                    imageType = u'ntfsclone'
                elif (head.find('partclone-image') != -1):
                    logger.notice(u"Image type is partclone")
                    imageType = u'partclone'

            except:
                if image:
                    image.close()
                raise

        if imageType not in (u'ntfsclone', u'partclone'):
            if isOpsi43:
                show_error("Unknown image type.")
            raise Exception(u"Unknown image type.")

        if disk.ldPreload:
            os.putenv("LD_PRELOAD", disk.ldPreload)
        logger.notice(u"Image type is %s" % imageType)
        if (imageType == u'partclone'):
            if useGpt:
                logger.info(u"Restoring partclone image '%s' to '%s'" % (imageFile, getPartitionDevicePath(diskDevice, partition)))
            else:
                logger.info(u"Restoring partclone image '%s' to '%s'" % (imageFile, disk.getPartition(partition)['device']) )

            #cmd = u'%s %s --source %s --overwrite %s' % \
            #                               (pipe, which('partclone.restore'), imageFile, disk.getPartition(partition)['device'])
            filename, file_extension = os.path.splitext(imageFile)
            # if extension = .img we restore from image file else we restore from loop mountable file
            if file_extension == '.img':
                if useGpt:
                    cmd = u'partclone.restore -C --force --source %s --output %s' % (imageFile, getPartitionDevicePath(diskDevice, partition))
                else:
                    cmd = u'partclone.restore -C --force --source %s --output %s' % (imageFile, disk.getPartition(partition)['device'])
            else:
                if useGpt:
                    cmd = u'partclone.%s --rescue -C --force --dev-to-dev --source %s --overwrite %s' % (filesystem, imageFile, getPartitionDevicePath(diskDevice, partition))
                else:
                    cmd = u'partclone.%s --rescue -C --force --dev-to-dev --source %s --overwrite %s' % (filesystem, imageFile, disk.getPartition(partition)['device'])
            logger.notice(u"Restoring partclone image using command: %s " % (cmd))
            if progressSubject:
                progressSubject.setEnd(100)
                progressSubject.setMessage(u"Scanning image")

            handle = execute(cmd, getHandle = True)
            done = False

            timeout = 0
            buf = [u'']
            lastMsg = u''
            started = False
            while not done:
                inp = handle.read(128)

                if inp:
                    inp = inp.decode("latin-1")
                    timeout = 0

                    b = inp.splitlines()
                    if inp.endswith(u'\n') or inp.endswith(u'\r'):
                        b.append(u'')

                    buf = [ buf[-1] + b[0] ] + b[1:]

                    for i in range( len(buf)-1 ):
                        try:
                            logger.debug(u" -->>> %s" % buf[i])
                        except:
                            pass
                        if ( buf[i].find(u'Partclone fail') != -1 ):
                            raise Exception(u"Failed: %s" % '\n'.join(buf))
                        if ( buf[i].find(u'Partclone successfully') != -1 ):
                            done = True
                        if not started:
                            if ( buf[i].count(':') == 1 ) and (buf[i].find('http:') == -1):
                                (k, v) = buf[i].split(':')
                                k = k.strip()
                                v = v.strip()
                                logger.info(u"Save image: %s: %s" % (k, v))
                                if progressSubject:
                                    progressSubject.setMessage(u"%s: %s" % (k, v))
                                if   (k.lower().find('file system') != -1):
                                    fs = v.lower()
                                elif (k.lower().find('used') != -1) or (k.lower().find('space in use') != -1):
                                    if progressSubject:
                                        progressSubject.setMessage(u"Restoring image")
                                    started = True
                                    continue
                        else:
                            #logger.debug(u" --> running:")
                            match = re.search('Completed:\s*([\d\.,]+)%', buf[i])
                            #logger.debug(u" --> match: %s" % match)
                            if match:
                                percent = int("%0.f" % float(match.group(1).replace(',','.')))
                                #logger.debug(u" --> detected percent progress: %d" % percent)
                                if isOpsi43:
                                    show_message("restore running", percent)
                                if progressSubject and (percent != progressSubject.getState()):
                                    logger.debug(u" --> setting percent progress: %d" % percent)
                                    #logger.debug(u" -->>> %s" % buf[i])
                                    progressSubject.setState(percent)

                    try:
                        lastMsg = buf[-2]
                    except IndexError:
                        logger.debug(u" -->>> %s" % len(buf))
                        lastMsg = buf[-1]
                    buf[:-1] = []

                elif (timeout >= 10000):
                    if progressSubject:
                        progressSubject.setMessage(u"Failed: %s" % lastMsg)
                    if isOpsi43:
                        show_error("Failed: %s" % lastMsg)
                    raise Exception(u"Failed: %s" % lastMsg)
                else:
                    timeout += 1
                    logger.debug(u" -->>> timeout %s" % timeout)
                    continue

            time.sleep(3)
            if handle: handle.close()

        if (fs == 'ntfs'):
            if not useGpt:
                disk.setNTFSPartitionStartSector(partition)
                if progressSubject:
                    progressSubject.setMessage(u"Resizing filesystem to partition size")
                disk.resizeFilesystem(partition, fs = u'ntfs')
            else:
                #disk.setNTFSPartitionStartSector(partition)
                # needed for GPT ?
                if progressSubject:
                    progressSubject.setMessage(u"Resizing filesystem to partition size")
                #execute("ntfsresize --no-progress-bar --force %s%d" % (diskDevice, partition) )
                # ntfsresize could only handle ntfs version 1.1
                os.system("ntfsresize --no-progress-bar --force %s" % (getPartitionDevicePath(diskDevice, partition)))
                #disk.resizeFilesystem(partition, fs = u'ntfs')

        if disk.ldPreload:
            os.unsetenv("LD_PRELOAD")

    except Exception as e:
        if not useGpt:
            for hook in hooks:
                hook.error_Harddisk_restoreImage(disk, partition, imageFile, progressSubject, e)
        raise

    for hook in hooks:
        if not useGpt:
            hook.post_Harddisk_restoreImage(disk, partition, imageFile, progressSubject)


useGpt = False
inUefi = inUefiMode()
if inUefi:
    useGpt = True

blockAlignment = True
useRaid1 = False

#method = productProperties.get('method', 'partclone-image-restore')
method = 'partclone-image-restore'
logger.notice(u"Restore method is %s" % method)

updateAndBackup = False
if (str(productProperties.get('update_and_backup', 'false')).lower() == 'true'):
    updateAndBackup = True
logger.notice(u"update_and_backup is %s" % updateAndBackup)

proxy = productProperties.get('proxy','')
no_proxy = productProperties.get('no_proxy','')

partgapM = 512

systemPartitionNumber = 1
systemPartitionLabel  = 'SYSTEM'
systemPartitionSizeM = 0

dataPartitionSizeM = 0
dataPartitionNumber = 0
dataPartitionLabel  = 'sysdata'

swapPartitionSizeM = 4000
winpePartitionNumber = 2
swapPartitionSize   = "4000M"
swapPartitionLabel  = "swap"
swapPartitionLetter = "X"

backupPartitionNumber = 3
backupPartitionCreate = True
backupPartitionLabel  = "opsidata"

if inUefi:
    systemPartitionNumber += 2
    winpePartitionNumber += 2
    backupPartitionNumber += 2

logger.notice("Partition numbers are now: system=%d, data=%s, swap=%d, backup=%d" % (systemPartitionNumber, dataPartitionNumber, winpePartitionNumber, backupPartitionNumber))

# get multiDiskMode and backup_partition_on_same_disk from opsi-local-image-prepare
multiDiskMode  = '0'
backup_partition_on_same_disk = True
if hasattr(backend, "productPropertyState_getValues"):
    # opsi >= 4.3
    for property_id, values in (
        backend.productPropertyState_getValues(product_ids=['opsi-local-image-prepare'], object_ids=[clientId], with_defaults=True)[clientId]['opsi-local-image-prepare']
        .items()
    ):
        logger.info("Found product property: %s=%r", property_id, values)
        if property_id == 'multi_disk_mode':
            multiDiskMode  = values[0]
        if property_id == 'backup_partition_on_same_disk':
            backup_partition_on_same_disk = values[0]
else:
    backend.backend_setOptions( { 'addProductPropertyStateDefaults': True } )
    pps = backend.productPropertyState_getObjects(objectId = clientId, productId = 'opsi-local-image-prepare',propertyId = 'multi_disk_mode')
    if pps:
        for state in pps:
            myvalues = state.getValues()
            for key in myvalues:
                multiDiskMode  = key
    pps = backend.productPropertyState_getObjects(objectId = clientId, productId = 'opsi-local-image-prepare',propertyId = 'backup_partition_on_same_disk')
    if pps:
        for state in pps:
            myvalues = state.getValues()
            for key in myvalues:
                backup_partition_on_same_disk  = key
    backend.backend_setOptions( { 'addProductPropertyStateDefaults': False } )
    
logger.notice("multiDiskMode: %s ; backup_partition_on_same_disk: %s" % (multiDiskMode,backup_partition_on_same_disk))

logger.notice(u"Swapoff all to avoid disk in use .....")
execute("swapoff -a -v")
# Get harddisks
disks = getHarddisks()
disks = sorted(disks, key=lambda disk: disk.device)
# Get harddisks
disks = getHarddisks()
disks = sorted(disks, key=lambda disk: disk.device)
showDiskOrder(disks)
logger.notice(u"Normally we using disk 0")
diskindex = getDiskIndexFromMultiDiskMode(multiDiskMode)
disk = disks[diskindex]
logger.notice(u"Using disk device %s" % (disk.device))
# check if (backup_partition_on_same_disk == False) and only one disk, what would make no sense
# and if so fall back to (backup_partition_on_same_disk == true)
backupdiskindex, backup_partition_on_same_disk = getBackupdiskindexFromBackupPartitionOnSameDisk(backup_partition_on_same_disk)
logger.notice(u"backup_partition_on_same_disk: %s - backupdiskindex %s" % (backup_partition_on_same_disk,backupdiskindex))
if (backup_partition_on_same_disk == False):
    # so we have a backup disk 
    backupPartitionNumber = 1

logger.notice("Partition numbers are now: system=%d, data=%s, swap=%d, backup=%d" % (systemPartitionNumber, dataPartitionNumber, winpePartitionNumber, backupPartitionNumber))

if not useGpt:
    # Display partition information
    partitions = disk.getPartitions()
    for p in partitions:
        fs   = p.get('fs', 'unknown')
        mytype = p.get('type', 'unknown')
        ui.getMessageBox().addText( u"Partition %s: Start %s MB, End %s MB, Filesystem %s, Type %s.\n" \
                                % (     p['number'],
                                        p['start']/(1000*1000),
                                        p['end']/(1000*1000),
                                        fs,
                                        mytype) )
else:
    execute('/usr/sbin/sgdisk -p %s' % (disk.device))

imageFileOrg = productProperties.get('imagefile', '')
scriptMessageSubject.setMessage( u"Will restore image: %s " % imageFileOrg)
logger.notice( u"Will restore image: %s " % imageFileOrg)
if imageFileOrg == '':
    if isOpsi43:
        show_error("No imagefile given - aborting")
    raise Exception(u"No imagefile given - aborting")
imageFile = re.sub("[\s\<\>\?\"\:\|\\\/\*]", "_",imageFileOrg)
logger.notice(u"Original image file name %s changed to internal image file name %s" % (imageFileOrg, imageFile))

setupAfterRestore = productPropertyValues.get('setup_after_restore', '')
if setupAfterRestore:
    for setupProduct in setupAfterRestore:
        logger.notice(u"Product to switch to setup after restore %s" % (setupProduct))


imageDir = '/mntimg'
prodDir = '/mntprod'
loopDir = '/mntloop'
goOn = True

checkOliPartitions(diskindex,backupdiskindex,useGpt,backup_partition_on_same_disk)
backupdisk = disks[backupdiskindex]
checkCreateDir(imageDir)
try:
    testPartitionEx(backupdisk.device, backupPartitionNumber, useGpt)
except:
    if isOpsi43:
        show_error("Could not find partition %d on disk: %s . - Please restart this product (opsi-local-image-prepare)" % (backupPartitionNumber,backupdisk.device))
    raise Exception(u"Could not find partition %d on disk: %s . - Please restart this product (opsi-local-image-prepare)" % (backupPartitionNumber,backupdisk.device))

partitionLabel = partitionHasLabel(backupdisk.device, backupPartitionNumber)
if (partitionLabel == backupPartitionLabel):
    logger.notice(u"%s%s exists as expected - fine" % (backupdisk.device, backupPartitionNumber) )
else:
    logger.critical(u"%s%s exists not as expected with label %s - not ok - giving up" % (backupdisk.device, backupPartitionNumber, backupPartitionLabel) )
    scriptMessageSubject.setMessage(u"%s%s exists not as expected with label %s - not ok - giving up" % (backupdisk.device, backupPartitionNumber, backupPartitionLabel) )
    if isOpsi43:
        show_error(u"Wrong partition table - Please run the opsi product 'opsi-local-image-pre pare' to prepare this disk!")
    raise Exception(u"Wrong partition table - Please run the opsi product 'opsi-local-image-prepare' to prepare this disk ! ")

# check if backup partition is in use
checkClearMountpoint(imageDir, backupdisk.device, backupPartitionNumber)
checkClearPartition(imageDir, backupdisk.device, backupPartitionNumber)
checkMount(imageDir, backupdisk.device, backupPartitionNumber)

if goOn:
    if not os.path.exists(imageDir):
        logger.critical(u"Could not find image base directory: %s - giving up" % imageDir)
        scriptMessageSubject.setMessage(u"Could not find image base directory: %s - giving up" % imageDir)
        if isOpsi43:
            show_error("Could not find image base directory: %s." % imageFileOrg)
        raise Exception(u"Could not find image base directory: %s." % imageFileOrg)
    if not os.path.exists("%s/%s" % (imageDir, imageFile)):
        logger.critical(u"Could not find expected image directory: %s/%s - giving up" % (imageDir, imageFile))
        scriptMessageSubject.setMessage(u"Could not find expected image directory: %s/%s - giving up" % (imageDir, imageFile))
        if isOpsi43:
            show_error("Could not find  given image directory for: %s." % imageFileOrg)
        raise Exception(u"Could not find  given image directory for: %s." % imageFileOrg)
    realimagefile = "%s/%s/%s" % (imageDir, imageFile, imageFile)
    if not os.path.exists("%s" % (realimagefile)):
        realimagefile = "%s/%s/%s.img" % (imageDir, imageFile, imageFile)
    logger.notice(u"Will restore image file: %s" % (realimagefile))

    # read meta data file
    iniFile = imageDir+'/'+imageFile+'/img.ini'
    if not os.path.exists(iniFile):
        logger.critical(u"Could not find meta data file: %s - giving up.." % (iniFile) )
        if isOpsi43:
            show_error("Could not find meta data file: %s - giving up.." % (iniFile) )
        raise Exception(u"Could not find meta data file: %s - giving up.." % (iniFile) )
    else:
        # Wenn Datei existiert parse Datei und gebe ConfigParser Objekt zurueck
        iniFileHandler = IniFile(iniFile)
        config = iniFileHandler.parse()
        if not config.has_section("%s" % imageFile.lower()):
            logger.error(u"Could not find section in meta data file: %s - bad.." % (iniFile) )
            if isOpsi43:
                show_error("Could not find section in meta data file: %s - bad.." % (iniFile))
            raise Exception(u"Could not find section in meta data file: %s - bad.." % (iniFile) )
        else:
            # read data
            filesystem = config.get(imageFile.lower(),"filesystem")
            logger.notice(u"Reading image meta data: filesystem: %s " % (filesystem) )
            opsiproduct = config.get(imageFile.lower(),"opsiproduct")
            logger.notice(u"Reading image meta data: opsiproduct: %s " % (opsiproduct) )
            imagedate = config.get(imageFile.lower(),"date")
            logger.notice(u"Reading image meta data: imagedate: %s " % (imagedate) )
            AveRate = config.get(imageFile.lower(),"AveRate")
            logger.notice(u"Reading image meta data: AveRate: %s " % (AveRate) )
            gptOnImage = config.get(imageFile.lower(),"GPT")
            logger.notice(u"Reading image meta data: GPT: %s " % (gptOnImage) )
            try:
                uefiFirstBootLabel = config.get(imageFile.lower(),"uefiFirstBootLabel")
                logger.notice(u"Reading image meta data: uefiFirstBootLabel: %s " % (uefiFirstBootLabel) )
                uefiFirstBootPartition = config.get(imageFile.lower(),"uefiFirstBootPartition")
                logger.notice(u"Reading image meta data: uefiFirstBootPartition: %s " % (uefiFirstBootPartition) )
                uefiFirstBootLoader = config.get(imageFile.lower(),"uefiFirstBootLoader")
                logger.notice(u"Reading image meta data: uefiFirstBootLoader: %s " % (uefiFirstBootLoader) )
            except:
                logger.notice(u"Reading image meta data: No uefiFirstBoot entries found")
                uefiFirstBootLabel = ''
                uefiFirstBootPartition = 0
                uefiFirstBootLoader = ''


# "partclone-image-restore"
if (method == "partclone-image-restore"):
    # check windows partition
    freespaceM = 0
    if not useGpt:
        try:
            SizeP = int(float(disk.getPartition(systemPartitionNumber).get("size"))/(1024*1024))
            logger.notice(u"Existing system partition size = %s" % SizeP)
        except:
            logger.notice(u"system partition (%d) seems not to exist" % (systemPartitionNumber))
            if isOpsi43:
                show_error("system partition (%d) seems not to exist" % (systemPartitionNumber))
            raise Exception(u"system partition (%d) seems not to exist" % (systemPartitionNumber))
        try:
            SizePe = int(float(disk.getPartition(winpePartitionNumber).get("size"))/(1024*1024))
            logger.notice(u"Existing winPE partition size = %s" % SizePe)
        except:
            logger.notice(u"swap partition (%d) seems not to exist" % (winpePartitionNumber))
            if isOpsi43:
                show_error("swap partition (%d) seems not to exist" % (winpePartitionNumber))
            raise Exception(u"swap partition (%d) seems not to exist" % (winpePartitionNumber))
    else:
        execute('/usr/sbin/sgdisk -p %s' % (disk.device))

    os.system('partprobe %s' % (disk.device))
    if not useGpt:
        partitions = disk.getPartitions()
    else:
        execute('/usr/sbin/sgdisk -p %s' % (disk.device))

    # change partition id and boot flag
    if useGpt:
        if (filesystem == 'ntfs'):
            logger.notice(u"change id of partition %s %d to 0700" % (disk.device,systemPartitionNumber))
            execute('/usr/sbin/sgdisk -t %d:0700 -p %s' % (systemPartitionNumber, disk.device))
            logger.notice(u"change id of partition %s %d to 1200" % (disk.device,winpePartitionNumber))
            execute('/usr/sbin/sgdisk -t %d:0700 -p %s' % (winpePartitionNumber, disk.device))
            # do never delete opsitempwinpe on a oli client (at least until we are able to recreate it in windows - do 18.1.22)
            # deleteAllEntriesFromEFIBootmgr('opsitempwinpe')
            removeFromUefiBootOrder('opsitempwinpe')
            removeFromUefiBootNext('opsitempwinpe')
            deleteAllEntriesFromEFIBootmgr('ubuntu')
        if (filesystem == 'ext4'):
            logger.notice(u"change id of partition %s %d to 8300" % (disk.device,systemPartitionNumber))
            execute('/usr/sbin/sgdisk -t %d:8300 -p %s' % (systemPartitionNumber, disk.device))
            logger.notice(u"change id of partition %s %d to 8200" % (disk.device,winpePartitionNumber))
            execute('/usr/sbin/sgdisk -t %d:8200 -p %s' % (winpePartitionNumber, disk.device))
    else:
        if (filesystem == 'ntfs'):
            logger.notice(u"change id of partition %s %d to 07" % (disk.device,systemPartitionNumber))
            execute('sfdisk --change-id %s %s 07' % (disk.device, systemPartitionNumber))
            logger.notice(u"change id of partition %s %d to 12" % (disk.device,winpePartitionNumber))
            execute('sfdisk --change-id %s %s 12' % (disk.device, winpePartitionNumber))
        if (filesystem == 'ext4'):
            logger.notice(u"change id of partition %s %d to 83" % (disk.device,systemPartitionNumber))
            execute('sfdisk --change-id %s %s 83' % (disk.device, systemPartitionNumber))
            logger.notice(u"change id of partition %s %d to 82" % (disk.device,winpePartitionNumber))
            execute('sfdisk --change-id %s %s 82' % (disk.device, winpePartitionNumber))
        logger.notice(u"Activate (set boot flag) on partition %s%d" % (disk.device,systemPartitionNumber))
        execute('sfdisk -A %s %d' % (disk.device,systemPartitionNumber))
        os.system('partprobe %s' % (disk.device))
        #exitcode = os.system(u'sfdisk --re-read %s' % disk.device)
        #if (0 != int(exitcode)):
        #       logger.notice(u"sfdisk --re-read %s failed - wait and retry" % (disk.device))
        #       execute("sleep 1")
        #       exitcode = os.system(u'sfdisk --re-read %s' % disk.device)
        #       if (0 != int(exitcode)):
        #               logger.notice(u"sfdisk --re-read %s failed - wait and retry" % (disk.device))
        #               execute("sleep 5")
        #               exitcode = os.system(u'sfdisk --re-read %s' % disk.device)
        #               if (0 != int(exitcode)):
        #                       logger.notice(u"sfdisk --re-read %s failed - giving up" % (disk.device))
        execute('sfdisk -l')

    def _getProgressSubject():
        from OPSI.Util.Message import ProgressSubject
        return ProgressSubject(id=u'backup_restore', title=u'Restoring Backup', end=100, fireAlways=True)

    try:
        if progressSubject is None:
            progressSubject = _getProgressSubject()
    except NameError:
        # No object defined.
        progressSubject = _getProgressSubject()

    backupProgressBox = ui.createProgressBox(width=60, height=20, total=100, title='Restore from Backup', text=u'')
    logger.debug('Created Progress box: {0}'.format(backupProgressBox))
    progressSubject.attachObserver(backupProgressBox)

    # restore
    #disk.restoreImage( partition = int(partition), imageFile = u"%s/%s/%s" % (imageDir, imageFile, imageFile) )
    restoreImage(diskDevice = disk.device, partition = int(systemPartitionNumber), imageType = "partclone", filesystem = filesystem, imageFile = u"%s" % (realimagefile), useGpt = useGpt, progressSubject=progressSubject)

    progressSubject.detachObserver(backupProgressBox)
    backupProgressBox.hide()

if not goOn:
    if isOpsi43:
        show_error("Failed - Please check log file")
    raise Exception(_(u'Failed - Please check log file'))

# restore bootloader
if (filesystem == 'ext4'):
    logger.notice(u"We restored ext4 filesystem and assume that we restored ubuntu")
    target = '/mntpostrestore'
    checkCreateDir(target)
    try:
        testPartitionEx(disk.device, systemPartitionNumber, useGpt)
    except:
        if isOpsi43:
            show_error("Could not find partition %d. - Please restart this product (opsi-local-image-prepare)" % systemPartitionNumber)
        raise Exception(u"Could not find partition %d. - Please restart this product (opsi-local-image-prepare)" % systemPartitionNumber)

    checkClearMountpoint(target, disk.device, systemPartitionNumber)
    checkClearPartition(target, disk.device, systemPartitionNumber)
    checkMount(target, disk.device, systemPartitionNumber, "ext4")

    execute("mount -o bind /proc %s/proc" % target)
    execute("mount -o bind /dev %s/dev" % target)
    execute("mount -o bind /sys %s/sys" % target)
    os.putenv("LD_PRELOAD", "")

    logger.notice(u"looking for %s/etc/debian_version" % target)
    if os.path.isfile("%s/etc/debian_version" % target):
        cmds = []
        cmds.append("export DEBIAN_PRIORITY=critical")
        cmds.append("export DEBIAN_FRONTEND=noninteractive")
        cmds.append("apt-get update")
        if inUefi:
            #http://wiki.ubuntuusers.de/GRUB_2/Reparatur
            cmds.append("mkdir -p /boot/efi")
            cmds.append("mount %s /boot/efi" % (getPartitionDevicePath(disk.device, 1)))
            cmds.append("mount")
            cmds.append("apt-get --yes --force-yes install grub-efi")
            cmds.append("grub-install %s" % (disk.device))
            cmds.append("grub-install --recheck %s" % (disk.device))
            cmds.append("update-grub ")
            cmds.append("umount %s" % (getPartitionDevicePath(disk.device, 1)))
            cmds.append("mount")
        else:
            cmds.append("apt-get --yes --force-yes install grub-pc")
            cmds.append("dpkg-reconfigure grub-pc")
            cmds.append("update-grub")
            cmds.append("grub-install --recheck %s" % disk.device)


        scriptMessageSubject.setMessage("Restore bootloader for Ubuntu...")
        logger.notice(u"Restore bootloader ...")
        doChrooted( target, cmds, proxy, no_proxy )
        cmds = []

    execute(" ls %s/" % target)
    execute(" ls %s/etc/" % target)
    logger.notice(u"looking for %s/etc/SuSE-release" % target)
    if os.path.isfile("%s/etc/SuSE-release" % target):
        cmds = []
        if inUefi:
            #http://wiki.ubuntuusers.de/GRUB_2/Reparatur
            cmds.append("mkdir -p /boot/efi")
            cmds.append("mount %s /boot/efi" % (getPartitionDevicePath(disk.device, 1)))
            cmds.append("mount")
            cmds.append("update-bootloader --refresh %s" % (getPartitionDevicePath(disk.device, 2)))
            cmds.append("grub2-mkconfig -o /boot/grub2/grub.cfg")
            cmds.append("grub2-install --recheck %s" % disk.device)
            cmds.append("umount %s" % (getPartitionDevicePath(disk.device, 1)))
            cmds.append("mount")
        else:
            cmds.append("update-bootloader --refresh %s" % (getPartitionDevicePath(disk.device, 2)))
            cmds.append("grub2-mkconfig -o /boot/grub2/grub.cfg")
            cmds.append("grub2-install --recheck %s" % disk.device)


        scriptMessageSubject.setMessage("Restore bootloader for opensuse...")
        logger.notice(u"Restore bootloader ...")
        doChrooted( target, cmds, proxy, no_proxy )
        cmds = []


    # Umount harddisk
    execute("umount %s/proc" % target)
    execute("umount %s/dev" % target)
    execute("umount %s/sys" % target)
    logger.notice(u"umount %s" % getPartitionDevicePath(disk.device, systemPartitionNumber))
    execute('/bin/umount %s' % getPartitionDevicePath(disk.device, systemPartitionNumber))

#if (int(partition) == 1):
if (filesystem == 'ntfs'):
    if not useGpt:
        # Write mbr
        #install-mbr -i n -p D -t 0 /dev/sda
        #disk.writeMasterBootRecord()
        logger.notice("Write Master Boot Record")
        execute("/usr/local/bin/ms-sys -v")
        if int(bootimageVersion) < 20181022:
            execute("/usr/local/bin/ms-sys --mbrnt60 %s" % (disk.device))
        else:
            execute("/usr/local/bin/ms-sys --mbr7 %s" % (disk.device))
    else:
        uefiFirstBootLabel = "Windows Boot Manager"
        if uefiFirstBootLabel != '':
            #deleteAllEntriesFromEFIBootmgr(uefiFirstBootLabel)
            logger.notice(u"setting %s to the top of uefi boot order" % uefiFirstBootLabel)
            setFirstUefiBoot(uefiFirstBootLabel)
            #loaderlist = uefiFirstBootLoader.split('\\')
            #extloader = '\\\\\\\\'.join(loaderlist)
            #logger.notice('executing: /bin/efibootmgr -c --disk %s --part %s -L %s --loader %s' % (disk.device, uefiFirstBootPartition, uefiFirstBootLabel, extloader ))
            #execute('/bin/efibootmgr -c --disk %s --part %s -L "%s" --loader %s' % (disk.device, uefiFirstBootPartition, uefiFirstBootLabel, extloader ))
        #os.system('/bin/efibootmgr -n 9')
        execute('/bin/efibootmgr -v')

if (filesystem == 'ext4') or (filesystem == 'ntfs'):
    # write back productOnClients
    logger.notice(u"Restoring productOnClients data from restored image")
    pocFile = imageDir+'/'+imageFile+'/productOnClients.json'
    f = open(pocFile, "r")
    lines = f.readlines()
    f.close()

    if lines:
        productsToConserve = ""
        try:
            productsToConserve = backend.productOnClient_getObjects(clientId=clientId,productId="opsi-local-image-wim-capture")
        except:
            logger.notice("No product opsi-local-image-wim-capture found")
        productOnClients = fromJson(lines[0])
        productsToDelete = backend.productOnClient_getObjects(clientId=clientId)
        if productsToDelete:
            backend.productOnClient_deleteObjects(productsToDelete)
        if productOnClients:
            backend.productOnClient_updateObjects(productOnClients)
        if productsToConserve:
            backend.productOnClient_updateObjects(productsToConserve)

    if updateAndBackup:
        logger.notice(u"updateAndBackup set - setting all outdated products to setup")
        # set all outdated products to setup
        clientIds = []
        clientIds.append(clientId)
        actionRequest = 'setup'
        # Get depot to client assignment
        depotToClients = {}
        myDepot = 'unknown'
        for clientToDepot in backend.configState_getClientToDepotserver(clientIds = clientIds):
            if "clientToDepot['depotId']" not in depotToClients:
                depotToClients[clientToDepot['depotId']] = []
            depotToClients[clientToDepot['depotId']].append(clientToDepot['clientId'])
            myDepot = clientToDepot['depotId']
        logger.notice(u"My depot seems to be %s" % myDepot)
        # get outdated products for client and depot
        updateProductOnClients = []
        myProductOnClients = backend.productOnClient_getObjects(clientId = clientId, productType = u'LocalbootProduct', installationStatus = u'installed')
        myProdctsOnDepot = backend.productOnDepot_getObjects(productType = 'LocalbootProduct', depotId = myDepot)
        for productOnClient in myProductOnClients:
            for productOnDepot in myProdctsOnDepot:
                if productOnClient.productId == productOnDepot.productId:
                    if (productOnClient.productVersion != productOnDepot.productVersion) or (productOnClient.packageVersion != productOnDepot.packageVersion):
                        logger.notice(u"Found product %s on client in version %s-%s and on depot %s-%s" % (productOnClient.productId,productOnClient.productVersion,productOnClient.packageVersion, productOnDepot.productVersion, productOnDepot.packageVersion ))
                        productOnClient.setActionRequest(actionRequest)
                        updateProductOnClients.append(productOnClient)
        backend.productOnClient_updateObjects(updateProductOnClients)
        # set opsi-local-image-backup-starter to once
        backend.setProductActionRequest(productId='opsi-local-image-backup-starter', clientId=clientId, actionRequest='once')

    if setupAfterRestore:
        for setupProduct in setupAfterRestore:
            if setupProduct != '':
                backend.setProductActionRequest(productId=setupProduct, clientId=clientId, actionRequest='setup')

if goOn:
    # write meta info to masterlog
    logstr = time.asctime( time.localtime(time.time()) )
    logstr = logstr + ', product=%s, filesystem=%s, given image file name=%s, used image file name=%s, method=%s, useGpt=%s' % (productId, filesystem, imageFileOrg, imageFile, method, useGpt)
    myvalues = []
    if hasattr(backend, "productPropertyState_getValues"):
        myvalues = backend.productPropertyState_getValues(product_ids=['opsi-local-image-restore'], object_ids=[clientId], property_ids=['imagefiles_list'], with_defaults=True)[clientId]['opsi-local-image-restore']['imagefiles_list']
    else:
        pps = backend.productPropertyState_getObjects(objectId = clientId, productId = 'opsi-local-image-restore',propertyId = 'imagefiles_list')
        if pps:
            for state in pps:
                myvalues = state.getValues()
    logstr = logstr + ', imageFiles_list=%s' % myvalues
    logFile = imageDir+'/master.log'
    if not os.path.exists(logFile):
        f = open(logFile, 'w' )
    else:
        f = open(logFile, 'a' )
    f.write(logstr)
    f.close()
    # show
    logger.notice(u"Here comes the master log:")
    execute(u'cat %s' % (logFile) )
    logger.notice(u"Here comes the backup partition usage:")
    execute(u"du -hs %s/*" % imageDir)

    # Set product installation status
    productOnClient.setInstallationStatus('installed')
    productOnClient.setActionProgress('')
    productOnClient.setActionResult('successful')
    productOnClient.setLastAction('setup')
else:
    # Set product installation status
    productOnClient.setInstallationStatus('unknown')
    productOnClient.setActionProgress('')
    productOnClient.setActionResult('failed')
    productOnClient.setLastAction('setup')

#repair imagefiles_list
myvalues = []
if hasattr(backend, "productPropertyState_getValues"):
    myvalues = backend.productPropertyState_getValues(product_ids=['opsi-local-image-restore'], object_ids=[clientId], property_ids=['imagefiles_list'], with_defaults=True)[clientId]['opsi-local-image-restore']['imagefiles_list']
else:
    pps = backend.productPropertyState_getObjects(objectId = clientId, productId = productId ,propertyId = 'imagefiles_list')
    if pps:
        for state in pps:
            myvalues = state.getValues()
for key in myvalues:
    if u',' in key:
        # This is the workaround for a bug where the list of
        # images will become a comma-seperated string instead of
        # a list of strings.
        for key_part in key.split(u','):
            if not (key_part in myvalues):
                myvalues.append(key_part)
        myvalues.remove(key)
backend.productPropertyState_create(productId = productId , propertyId = 'imagefiles_list', objectId = clientId, values = myvalues)


# Reboot machine
reboot()
